@@ -26,13 +26,13 @@
 #include <unistd.h>
 #include <dirent.h>
 #include <libgen.h>
-#include "ctree.h"
-#include "transaction.h"
-#include "utils.h"
+#include "kernel-shared/ctree.h"
+#include "kernel-shared/transaction.h"
+#include "common/utils.h"
 #include "ioctl.h"
 #include <uuid/uuid.h>
 #include "btrfs-list.h"
-#include "rbtree-utils.h"
+#include "common/rbtree-utils.h"
 
 #define BTRFS_LIST_NFILTERS_INCREASE	(2 * BTRFS_LIST_FILTER_MAX)
 #define BTRFS_LIST_NCOMPS_INCREASE	(2 * BTRFS_LIST_COMP_MAX)
