@@ -19,6 +19,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
+#include <sys/vfs.h>
 #include <errno.h>
 #include <stdarg.h>
 #include <getopt.h>
@@ -31,10 +32,12 @@
 #include "cmds/filesystem-usage.h"
 #include "cmds/commands.h"
 #include "kernel-shared/disk-io.h"
-
+#include "common/open-utils.h"
+#include "common/units.h"
 #include "version.h"
 #include "common/help.h"
 #include "common/device-utils.h"
+#include "common/open-utils.h"
 
 /*
  * Add the chunk info to the chunk_info list
@@ -370,7 +373,7 @@ static void get_raid56_space_info(struct btrfs_ioctl_space_args *sargs,
 			*max_data_ratio = rt;
 
 		/*
-		 * size is the total disk(s) space occuped by a chunk
+		 * size is the total disk(s) space occupied by a chunk
 		 * the product of 'size' and  '*_ratio' is "in average"
 		 * the disk(s) space used by the data
 		 */
@@ -428,8 +431,11 @@ static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 	u64 l_global_reserve_used = 0;
 	u64 free_estimated = 0;
 	u64 free_min = 0;
+	u64 zone_unusable = 0;
 	double max_data_ratio = 1.0;
 	int mixed = 0;
+	struct statfs statfs_buf;
+	struct btrfs_ioctl_feature_flags feature_flags;
 
 	sargs = load_space_info(fd, path);
 	if (!sargs) {
@@ -494,6 +500,16 @@ static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 		if ((flags & (BTRFS_BLOCK_GROUP_DATA | BTRFS_BLOCK_GROUP_METADATA))
 		    == (BTRFS_BLOCK_GROUP_DATA | BTRFS_BLOCK_GROUP_METADATA)) {
 			mixed = 1;
+		} else {
+			/*
+			 * As mixed mode is not supported in zoned mode, this
+			 * will account for all profile types
+			 */
+			u64 tmp;
+
+			tmp = device_get_zone_unusable(fd, flags);
+			if (tmp != DEVICE_ZONE_UNUSABLE_UNKNOWN)
+				zone_unusable += tmp;
 		}
 		if (flags & BTRFS_BLOCK_GROUP_DATA) {
 			r_data_used += sargs->spaces[i].used_bytes * ratio;
@@ -556,6 +572,13 @@ static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 	if (unit_mode != UNITS_HUMAN)
 		width = 18;
 
+	ret = statfs(path, &statfs_buf);
+	if (ret) {
+		warning("cannot get space info with statfs() on '%s': %m", path);
+		memset(&statfs_buf, 0, sizeof(statfs_buf));
+		ret = 0;
+	}
+
 	printf("Overall:\n");
 
 	printf("    Device size:\t\t%*s\n", width,
@@ -568,10 +591,17 @@ static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 		pretty_size_mode(r_total_missing, unit_mode));
 	printf("    Used:\t\t\t%*s\n", width,
 		pretty_size_mode(r_total_used, unit_mode));
+	ret = ioctl(fd, BTRFS_IOC_GET_FEATURES, &feature_flags);
+	if (ret == 0 && (feature_flags.incompat_flags & BTRFS_FEATURE_INCOMPAT_ZONED)) {
+		printf("    Device zone unusable:\t%*s\n", width,
+			pretty_size_mode(zone_unusable, unit_mode));
+	}
 	printf("    Free (estimated):\t\t%*s\t(",
 		width,
 		pretty_size_mode(free_estimated, unit_mode));
 	printf("min: %s)\n", pretty_size_mode(free_min, unit_mode));
+	printf("    Free (statfs, df):\t\t%*s\n", width,
+		pretty_size_mode(statfs_buf.f_bavail * statfs_buf.f_bsize, unit_mode));
 	printf("    Data ratio:\t\t\t%*.2f\n",
 		width, data_ratio);
 	printf("    Metadata ratio:\t\t%*.2f\n",
@@ -701,7 +731,7 @@ static int load_device_info(int fd, struct device_info **device_info_ptr,
 		} else {
 			strcpy(info[ndevs].path, (char *)dev_info.path);
 			info[ndevs].device_size =
-				get_partition_size((char *)dev_info.path);
+				device_get_partition_size((const char *)dev_info.path);
 		}
 		info[ndevs].size = dev_info.total_bytes;
 		++ndevs;
@@ -882,7 +912,7 @@ static void _cmd_filesystem_usage_tabular(unsigned unit_mode,
 			col++;
 		}
 
-		unused = get_partition_size(device_info_ptr[i].path)
+		unused = device_get_partition_size(device_info_ptr[i].path)
 				- total_allocated;
 
 		table_printf(matrix, unallocated_col, vhdr_skip + i, ">%s",
